from enum import IntFlag


class Intents(IntFlag):
    """
    Intents, otherwise known as "Gateway intents" are a set of
    bitwise values that help control the computational burden
    received by a Gateway client.

    ---

    Some intents in this class are documented as being "privileged,"
    a special intent that must be approved before receiving access
    to its respective data. In order to do this, the application in
    question (e.g. a bot) must have the intent enabled through
    the Developer Portal settings of said application.

    The following intents have this label:

    - `GUILD_PRESENCES` - event(s) relating to a guild member's presence.
    - `GUILD_MEMBERS` - event(s) relating to a guild member.
    - `MESSAGE_CONTENT` - event(s) relating to the `content`, `embeds`, `attachments` or `components` fields of an object with said limitation.

    Additionally, you are able to access all of these with `PRIVILEGED`.

    ---

    A bot should never have intents that will ultimately
    be discarded or unused. Please see all intents given. `ALL`
    may be used but must be acknowledged for carrying every
    intent given by this class.
    """

    GUILDS = 1 << 0
    """
    `GUILDS` controls a majority of events shown for any guild a bot is in.
    The following events are gated behind this intent:

    - `GUILD_CREATE` - upon joining a guild.
    - `GUILD_UPDATE` - upon a guild updating data.
    - `GUILD_DELETE` - upon leaving a guild.
    - `GUILD_ROLE_CREATE` - upon a role in a guild being created.
    - `GUILD_ROLE_UPDATE` - upon a role's data in a guild being updated.
    - `GUILD_ROLE_DELETE` - upon a rolei n a guild being deleted.
    - `CHANNEL_CREATE` - upon a channel in a guild being created.
    - `CHANNEL_UPDATE` - upon a channel's data in a guild being updated.
    - `CHANNEL_DELETE` - upon a channel in a guild being deleted.
    - `CHANNEL_PINS_UPDATE` - upon a pin in a channel being created, updated or deleted.
    - `THREAD_CREATE` - upon a thread in a guild being created, or the bot added to one.
    - `THREAD_UPDATE` - upon a thread in a guild being updated. This event sometimes coincides with `THREAD_CREATE`.
    - `THREAD_DELETE` - upon a thread in a guild being deleted.
    - `THREAD_LIST_SYNC` - upon a thread in a guild updating its member list. This event ultimately follows `THREAD_UPDATE`.
    - `THREAD_MEMBER_UPDATE` - upon a member in a guild thread being updated.
    - `THREAD_MEMBERS_UPDATE` - upon a set of members in a guild thread being updated. This event ultimately follows `THREAD_UPDATE`.
    - `STAGE_INSTANCE_CREATE` - upon a stage event in a guild being created.
    - `STAGE_INSTANCE_UPDATE` - upon a stage event's data in a guild being updated.
    - `STAGE_INSTANCE_DELETE` - upon a stage event in a guild being deleted.
    """
    GUILD_MEMBERS = 1 << 1
    """
    This intent is considered to be "privileged." Please see the documentation
    on the main class (`Intents`) to learn more.

    `GUILD_MEMBERS` controls events for a guild member's data being shown.
    The following events are gated behind this intent:

    - `GUILD_MEMBER_ADD` - upon a member being added to a guild.
    - `GUILD_MEMBER_UPDATE` - upon a member's data in a guild being updated.
    - `GUILD_MEMBER_REMOVE` - upon a member being removed from a guild.
    - `THREAD_MEMBERS_UPDATE` - Coinciding with the `GUILDS` intent. (see)
    """
    GUILD_BANS = 1 << 2
    """
    `GUILD_BANS` controls events for a guild member's ban state.
    The following events are gated behind this intent:

    - `GUILD_BAN_ADD` - upon a member being banned from a guild. `GUILD_MEMBER_REMOVE` ultimately follows this.
    - `GUILD_BAN_REMOVE` - upon a member being unbanned from a guild.
    """
    GUILD_EMOJIS_AND_STICKERS = 1 << 3
    """
    `GUILD_EMOJIS_AND_STICKERS` controls events for emojis and stickers
    being sent. The following events are gated behind this intent:

    - `GUILD_EMOJIS_UPDATE` - upon an emoji or set being updated in a guild.
    - `GUILD_STICKERS_UPDATE` - upon a sticker or set being updated in a guild.
    """
    GUILD_INTEGRATIONS = 1 << 4
    """
    `GUILD_INTEGRATIONS` controls events for integrations.
    The following events are gated behind this intent:

    - `GUILD_INTEGRATIONS_UPDATE` - upon an integration or set being updated in a guild.
    - `INTEGRATION_CREATE` - upon an integration regardless of guild status being created.
    - `INTEGRATION_UPDATE` - upon an integration regardless of guild status being updated.
    - `INTEGRATION_DELETE` - upon an integration regardless of guild status being deleted.
    """
    GUILD_WEBHOOKS = 1 << 5
    """
    `GUILD_WEBHOOKS` controls an event for webhooks in a guild.
    The following event is gated behind this intent:

    - `WEBHOOKS_UPDATE` - upon a webhook in a guild being updated.
    """
    GUILD_INVITES = 1 << 6
    """
    `GUILD_INVITES` controls events for invites in a guild.
    The following events are gated behind this intent:

    - `INVITE_CREATE` - upon a guild invitation being created.
    - `INVITE_DELETE` - upon a guild invitation being deleted.
    """
    GUILD_VOICE_STATES = 1 << 7
    """
    `GUILD_VOICE_STATES` controls an event for a guild member's voice state.
    The following event is gated behind this intent:

    - `VOICE_STATE_UPDATE` - upon a guild member's voice state being updated.
    """
    GUILD_PRESENCES = 1 << 8
    """
    This intent is considered to be "privileged." Please see the documentation
    on the main class (`Intents`) to learn more.

    `GUILD_PRESENCES` controls an event for a guild member's presence.
    The following event is gated behind this intent:

    - `PRESENCE_UPDATE` - upon a guild member's presence being updated.
    """
    GUILD_MESSAGES = 1 << 9
    """
    `GUILD_MESSAGES` controls events for messages in a guild.
    The following events are gated behind this intent:

    - `MESSAGE_CREATE` - upon a message in a guild being created.
    - `MESSAGE_UPDATE` - upon a message in a guild being updated.
    - `MESSAGE_DELETE` - upon a message in a guild being deleted.
    - `MESSAGE_DELETE_BULK` - upon a set of messages in a guild being deleted.
    """
    GUILD_MESSAGE_REACTIONS = 1 << 10
    """
    `GUILD_MESSAGE_REACTIONS` controls events for reactions on a guild message.
    The following events are gated behind this intent:

    - `MESSAGE_REACTION_ADD` - upon a reaction on a guild message being added.
    - `MESSAGE_REACTION_REMOVE` - upon a reaction on a guild message being removed.
    - `MESSAGE_REMOVE_ALL` - upon a guild member with the "Manage Messages" permission removing all reactions on a guild message.
    - `MESSAGE_REMOVE_EMOJI` - ???
    """
    GUILD_MESSAGE_TYPING = 1 << 11
    """
    `GUILD_MESSAGE_TYPING` controls an event for a guild member typing.
    The following event is gated behind this intent:

    - `TYPING_START` - upon a guild member typing in a guild.
    """
    DIRECT_MESSAGES = 1 << 12
    """
    `DIRECT_MESSAGES` controls events for direct messages.
    The following events are the same as `GUILD_MESSAGES`, with the exception of:

    - `CHANNEL_PINS_UPDATE` - upon a set of pins in a guild channel being updated.
    """
    DIRECT_MESSAGE_REACTIONS = 1 << 13
    """
    `DIRECT_MESSAGE_REACTIONS` controls events for reactions on a direct message.
    The following events are the same as `GUILD_MESSAGE_REACTIONS`.
    """
    DIRECT_MESSAGE_TYPING = 1 << 14
    """
    `DIRECT_MESSAGE_TYPING` controls an event for a user typing.
    The following event is the same as `GUILD_MESSAGE_TYPING`.
    """
    MESSAGE_CONTENT = 1 << 15
    """
    This intent is considered to be "privileged." Please see the documentation
    on the main class (`Intents`) to learn more.

    `MESSAGE_CONTENT` controls data passed to these specific fields
    in objects when specified:
    - `content` - data of the exact contents given on a message.
    - `embeds` - an embed or set on a message.
    - `attachments` - an attachment or set on a message.
    - `components` - a component or set on a message.
    """
    GUILD_SCHEDULED_EVENTS = 1 << 16
    """
    `GUILD_SCHEDULED_EVENTS` controls events for scheduled events in a guild.
    The following events are gated behind this intent:

    - `GUILD_SCHEDULED_EVENT_CREATE` - upon a scheduled event in a guild being created.
    - `GUILD_SCHEDULED_EVENT_UPDATE` - upon a scheduled event in a guild being updated.
    - `GUILD_SCHEDULED_EVENT_DELETE` - upon a scheduled event in a guild being deleted.
    - `GUILD_SCHEDULED_EVENT_USER_ADD` - upon a guild member in a scheduled event being added.
    - `GUILD_SCHEDULED_EVENT_USER_REMOVE` - upon a guild member in a scheduled event being removed.
    """
    AUTO_MODERATION_CONFIGURATION = 1 << 20
    """
    `AUTO_MODERATION_CONFIGURATION` controls events for settings of a guild's AutoMod.
    The following events are gated behind this intent:

    - `AUTO_MODERATION_RULE_CREATE` - upon an automoderation rule being created.
    - `AUTO_MODERATION_RULE_UPDATE` - upon an automoderation rule being updated.
    - `AUTO_MODERATION_RULE_DELETE` - upon an automoderation rule being deleted.
    """
    AUTO_MODERATION_EXECUTION = 1 << 21
    """
    `AUTO_MODERATION_EXECUTION` controls an event for executing AutoMod rules.
    The following event is gated behind this intent:

    - `AUTO_MODERATION_ACTION_EXECUTION` - upon an automoderation rule or set being executed.
    """

    PRIVILEGED = GUILD_PRESENCES | GUILD_MEMBERS | MESSAGE_CONTENT
    """
    `PRIVILEGED` is the culmination of all privileged intents.
    The following intents are known to be "privileged:"

    - `GUILD_PRESENCES`
    - `GUILD_MEMBERS`
    - `MESSAGE_CONTENT`
    """
    NON_PRIVILEGED = (
        GUILDS
        | GUILD_BANS
        | GUILD_EMOJIS_AND_STICKERS
        | GUILD_INTEGRATIONS
        | GUILD_WEBHOOKS
        | GUILD_INVITES
        | GUILD_VOICE_STATES
        | GUILD_MESSAGES
        | GUILD_MESSAGE_TYPING
        | DIRECT_MESSAGES
        | DIRECT_MESSAGE_TYPING
        | DIRECT_MESSAGE_REACTIONS
        | GUILD_SCHEDULED_EVENTS
        | AUTO_MODERATION_CONFIGURATION
        | AUTO_MODERATION_EXECUTION
    )
    """
    `NON_PRIVILEGED` is the culmination of all intents that are not
    "privileged." See `PRIVILEGED` for the missing intents.
    """
    ALL = PRIVILEGED | NON_PRIVILEGED
    """
    `ALL` is the culmination of all intents of this class.

    If necessary, see `PRIVILEGED` and `NON_PRIVILEGED` for their inclusion.
    """



from datetime import datetime
from attrs import define, field

from ...client.resources.abc import Snowflake
from ...client.resources.guild import Member
from ...utils.converters import optional_c


@define()
class TypingStart:
    """
    Represents a `TYPING_START` event from Discord.

    ---

    Sent when a user starts typing in a channel.

    ---

    Attributes
    ----------
    channel_id : `Snowflake`
        The ID of the channel when typing occured.
    user_id : `Snowflake`
        The ID of the user who started typing.
    timestamp : `datetime.datetime`
        The timestamp of when the typing occured.
    guild_id : `Snowflake`, optional
        The ID of the guild when typing occured.

        This will only appear when a user is typing
        outside of a DM.
    member : `Member`, optional
        The member who started typing.

        This will only appear when a user is typing
        outside of a DM.
    """

    channel_id: str | Snowflake = field(converter=Snowflake)
    """The ID of the channel when typing occured."""
    user_id: str | Snowflake = field(converter=Snowflake)
    """The ID of the user who started typing."""
    timestamp: int | datetime = field(converter=datetime.fromtimestamp)
    """The timestamp of when the typing occured."""
    guild_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """
    The ID of the guild when typing occured.

    This will only appear when a user is typing
    outside of a DM.
    """
    member: dict | Member | None = field(converter=optional_c(Member), default=None)
    """
    The member who started typing.

    This will only appear when a user is typing
    outside of a DM.
    """



from enum import IntEnum, IntFlag
from attrs import define, field

from .abc import Snowflake, Object

__all__ = ("User", "UserFlags", "UserPremiumType")


class UserFlags(IntFlag):
    """User flags are a set of bitwise values that represent the public flags of a user."""

    STAFF = 1 << 0
    """This user is a Discord employee."""
    PARTNER = 1 << 1
    """This user is part of the Partnership program."""
    HYPESQUAD = 1 << 2
    """This user is part of the HypeSquad Events program."""
    BUG_HUNTER_LEVEL_1 = 1 << 3
    """This user has the first Bug Hunter level."""
    HYPESQUAD_ONLINE_HOUSE_1 = 1 << 6
    """This user is part of the HypeSquad Bravery house."""
    HYPESQUAD_ONLINE_HOUSE_2 = 1 << 7
    """This user is part of the HypeSquad Brilliance house."""
    HYPESQUAD_ONLINE_HOUSE_3 = 1 << 8
    """This user is part of the HypeSquad Balance house."""
    PREMIUM_EARLY_SUPPORTER = 1 << 9
    """This user is an early Nitro supporter."""
    TEAM_PSEUDO_USER = 1 << 10
    """This user represents a `Team` on Discord."""
    BUG_HUNTER_LEVEL_2 = 1 << 14
    """This user has the second Bug Hunter level."""
    VERIFIED_BOT = 1 << 16
    """This user is a bot whose verification passed."""
    VERIFIED_DEVELOPER = 1 << 17
    """This user is a verified bot developer."""
    CERTIFIED_MODERATOR = 1 << 18
    """This user is a certified moderator."""
    BOT_HTTP_INTERACTIONS = 1 << 19
    """This user is a bot that only intakes interaction endpoints."""


class UserPremiumType(IntEnum):
    """
    Represents the types of user premium levels on Discord.

    Constants
    ---------
    NONE
        This user has no premium subscription.
    NITRO_CLASSIC
        This user has the Nitro Classic subscription.
    NITRO
        This user has the Nitro Regular subscription.
    """

    NONE = 0
    """This user has no premium subscription."""
    NITRO_CLASSIC = 1
    """This user has the Nitro Classic subscription."""
    NITRO = 2
    """This user has the Nitro Regular subscription."""


@define(kw_only=True)
class User(Object):
    """
    Represents a user from Discord.

    ---

    Because a user's information can be accessed via. numerous methods,
    some information is marked as being optional, when in reality it
    cannot be accessed by conventional means. Only `verified` and `email`
    are `None` if either not set, or OAuth2 was used with the e-mail scope.

    ---

    Attributes
    ----------
    id : `Snowflake`
        The ID of the user.
    username : `str`
        The name of the user.
    discriminator : `str`
        The discriminator (4-digit tag) of the user.
    avatar : `str`, optional
        The hash of the user's avatar.
    bot : `bool`, optional
        Whether the user is a bot or not.
    system : `bool`, optional
        Whether the user is from the official Discord System or not.
    mfa_enabled : `bool`, optional
        Whether the user has 2FA (two-factor authentication) enabled
        or not.
    banner : `str`, optional
        The hash of the user's banner, if present.
    accent_color : `int`, optional
        The color of the user's banner, if present.
    locale : `str`, optional
        The user's selected locale, if present.
    verified : `bool`, optional
        Whether the user has a verified e-mail or not.
    email : `str`, optional
        The e-mail associated to the user's account, if present.
    flags : `UserFlags`, optional
        The public flags on the user's account, if present.
    premium_type : `int`, optional
        The type of Nitro subscription the user has.
    public_flags : `UserFlags`, optional
        The public flags on the user's account, if present.

    Methods
    -------
    tag : `str`
        The 4-digit tag of the user.
    mention : `str`
        The mentionable version of the user.
    """

    id: str | Snowflake = field(converter=Snowflake)
    """The ID of the user."""
    username: str = field()
    """The name of the user."""
    discriminator: str = field()
    """The discriminator (4-digit tag) of the user."""
    avatar: str | None = field(default=None)
    """The hash of the user's avatar."""
    bot: bool = field(default=False)
    """Whether the user is a bot or not."""
    system: bool = field(default=False)
    """Whether the user is from the official Discord System or not."""
    mfa_enabled: bool = field(default=False)
    """Whether the user has 2FA (two-factor authentication) enabled or not."""
    banner: str | None = field(default=None)
    """The hash of the user's banner, if present."""
    accent_color: int | None = field(default=None)
    """The color of the user's banner, if present."""
    locale: str | None = field(default=None)
    """The user's selected locale, if present."""
    verified: bool = field(default=True)
    """Whether the user has a verified e-mail or not."""
    email: str | None = field(default=None)
    """The e-mail associated to the user's account, if present."""
    flags: int | UserFlags = field(converter=UserFlags, default=0)
    """The public flags on the user's account, if present."""
    premium_type: int | UserPremiumType = field(converter=UserPremiumType, default=0)
    """The type of Nitro subscription the user has."""
    public_flags: int | UserFlags = field(converter=UserFlags, default=0)
    """The public flags on the user's account, if present."""

    @property
    def tag(self) -> str:
        """The 4-digit tag of the user."""
        return self.discriminator

    @property
    def mention(self) -> str:
        """The mentionable version of the user."""
        return f"<@{self.id}>"



from .user import User
from .guild import Guild, Member
from .channel import Channel
from ...utils.converters import optional_c, list_c
from .application import Application
from datetime import datetime

from attrs import define, field
from enum import IntEnum


class InviteTargetType(IntEnum):
    """
    Represents the targeted types of invites from Discord.

    Constants
    ---------
    STREAM
        The invite targets a stream.
    EMBEDDED_APPLICATION
        The invite targets an embedded application.
    """

    STREAM = 1
    """The invite targets a stream."""
    EMBEDDED_APPLICATION = 2
    """The invite targets an embedded application."""


@define()
class InviteStageInstance:
    """
    Represents an invite to a stage instance from Discord. This is deprecated.

    Attributes
    ----------
    members : `list[retux.Member]`
        The members speaking in the stage.
    participant_count : `int`
        The number of users in the stage.
    speaker_count : `int`
        The number of users speaking in the stage.
    topic : `str`
        The topic of the stage instance in-between 1-120 characters.
    """

    members: list[Member] = field(converter=list_c(Member))
    """The members speaking in the stage."""
    participant_count: int = field()
    """The number of users in the stage."""
    speaker_count: int = field()
    "The number of users speaking in the stage."
    topic: str = field()
    "The topic of the stage instance in-between 1-120 characters."


@define()
class Invite:
    """
    Represents an invite from Discord.

    Attributes
    ----------
    code : `str`
        The invite code.
    guild : `retux.Guild`, optional
        The guild this invite is for, if any.
    channel : `retux.Channel`, optional
        The channel this invite is for, if any.
    inviter : `retux.User`, optional
        The user who created the invite, if any.
    target_type : `InviteTargetType`, optional
        The type of target for a voice channel invite, if any.
    target_user : `retux.User`, optional
        The user whose stream to display for this voice channel stream invite, if any.
    target_application : `retux.Application`, optional
        The embedded application to open for this voice channel embedded application invite, if any.
    approximate_presence_count : `int`, optional
        Approximate count of online members, returned from the `Get Invites` endpoint
        when `with_counts=True`.
    approximate_member_count : `int`, optional
        Approximate count of total members, returned from the `Get Invites` endpoint
        when `with_counts=True`.
    expires_at : `datetime.datetime`, optional
        The expiration date of this invite, returned from the `Get Invites` endpoint
        when `with_expiration=True`.
    stage_instance : `InviteStageInstance`, optional
        Data of the instance if there is a public stage in the channel this invite is for. (deprecated)
    """

    code: str = field()
    """The invite code."""
    guild: dict | Guild | None = field(converter=optional_c(Guild), default=None)
    """The guild this invite is for, if any."""
    channel: dict | Channel | None = field(converter=optional_c(Channel), default=None)
    """The channel this invite is for, if any."""
    inviter: dict | User | None = field(converter=optional_c(User), default=None)
    """The user who created the invite, if any."""
    target_type: int | InviteTargetType | None = field(
        converter=optional_c(InviteTargetType), default=None
    )
    """The type of target for this voice channel invite, if any."""
    target_user: dict | User | None = field(converter=optional_c(User), default=None)
    """The user whose stream to display for this voice channel stream invite, if any."""
    target_application: dict | Application | None = field(
        converter=optional_c(Application), default=None
    )
    """The embedded application to open for this voice channel embedded application invite, if any"""
    approximate_presence_count: int | None = field(default=None)
    """Approximate count of online members, returned from the `Get Invites` endpoint when `with_counts=True`."""
    approximate_member_count: int | None = field(default=None)
    """Approximate count of total members, returned from the `Get Invites` endpoint when `with_counts=True`."""
    expires_at: str | datetime | None = field(
        converter=optional_c(datetime.fromisoformat), default=None
    )
    """The expiration date of this invite, returned from the `Get Invites` endpoint when `with_expiration=True`."""
    stage_instance: int | InviteStageInstance | None = field(
        converter=optional_c(InviteStageInstance), default=None
    )
    """Data of the instance if there is a public stage in the channel this invite is for. (deprecated)"""
    # guild_scheduled_event
    # """guild scheduled event data, only included if guild_scheduled_event_id contains a valid guild scheduled event id"""
    # TODO: Implement Guild Scheduled Event object



from typing import Any, Callable


def optional_c(converter: Callable) -> Callable[..., Any]:
    """
    Handles conversion for possibly optional/missing values.

    Attributes
    ----------
    converter : `typing.Callable`
        The callable method to convert to.

        This could be a method or a class
        that is alone or `attrs`-based.

    Returns
    -------
    `typing.Callable[..., typing.Any]`
        Any possible value. If the value of
        the field is found to be `None` or
        `MISSING`, it will skip any given
        conversions. Otherwise, it will returned
        the converted value from `converter`.
    """

    def optional_converter(val):
        if val is None:
            return val
        return converter(val)

    return optional_converter


def list_c(func: Any) -> Callable[..., list[Any]]:
    """
    Handles conversion for values nested inside of a list.

    ---

    This converter can be layered inside of `optional_c`,
    allowing for granular control over optional lists of
    values.

    ---

    Attributes
    ----------
    func : `typing.Any`
        The function or value to convert the list of.
        The function is needed as it's the presumed type
        of the converter when handled.

    Returns
    -------
    `list[typing.Any]`
        A list of values converted into what `func`
        has been provided with.
    """

    def inner(val):
        if val is not None:
            return [func(_) for _ in val]

    return inner



from typing import TypeVar, Union


__version__ = "0.0.3"
__api_version__ = "v10"
__api_url__ = f"https://discord.com/api/{__api_version__}"
__gateway_url__ = "wss://gateway.discord.gg/"
__repo_url__ = "https://github.com/i0bs/retux"


class MISSING:
    """
    A sentinel that represents an argument with a "missing" value.
    This is used deliberately to avoid `None` space confusion.
    """

    pass


_T = TypeVar("_T")
NotNeeded = Union[_T, MISSING]
"""
A type variable to work alongside `MISSING`. This should only
be used to help further indicate an optional argument where it
already points to said type.
"""



from os import path
from setuptools import find_packages, setup

PROJECT_NAME = "retux"
PROJECT_DESC = "A Discord API wrapper built with good intentions."


def read(fp):
    return open(path.join(path.dirname(__file__), fp)).read()


VERSION = "0.0.3"
README = read("README.md")
# REQUIREMENTS = open("requirements.txt", "r").read().strip().splitlines()
AUTHOR_NAME = "i0"
AUTHOR_EMAIL = "me@i0.gg"

setup(
    name=PROJECT_NAME,
    description=PROJECT_DESC,
    long_description=README,
    long_description_content_type="text/markdown",
    version=VERSION,
    author=AUTHOR_NAME,
    author_email=AUTHOR_EMAIL,
    maintainer=AUTHOR_NAME,
    maintainer_email=AUTHOR_EMAIL,
    url="https://github.com/i0bs/retux",
    license="AGPL-3.0",
    keywords="python discord discord-bot discord-api python3 discord-bots",
    packages=find_packages(),
    include_package_data=True,
    install_requires=["attrs", "cattrs", "httpx", "trio", "trio_websocket"],
    python_requires=">=3.10.0",
    classifiers=[
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Topic :: Internet",
        "Topic :: Utilities",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



from datetime import datetime
from enum import IntFlag
from typing import Any
from enum import IntEnum
from attrs import define, field

from .user import User, UserFlags, UserPremiumType
from .emoji import Emoji
from .sticker import Sticker
from .role import Role

from .abc import Object, Partial, Snowflake

from ...utils.converters import optional_c, list_c

__all__ = (
    "Guild",
    "UnavailableGuild",
    "GuildPreview",
    "GuildWidget",
    "WelcomeScreen",
    "WelcomeScreenChannel",
    "SystemChannelFlags",
    "GuildNSFWLevel",
    "ExplicitContentFilterLevel",
    "VerificationLevel",
    "Member",
    "GuildWidgetSettings",
)


@define(kw_only=True)
class WelcomeScreenChannel:
    """
    Represents a channel shown in the welcome screen of a guild from Discord.

    Attributes
    ----------
    channel_id : `Snowflake`
        The ID of the channel shown.
    description : `str`
        A description shown with the channel.
    emoji_id : `Snowflake`, optional
        The ID of the emoji if it isn't a unicode.
    emoji_name : `str`, optional
        The name of the emoji, if present.
    """

    channel_id: str | Snowflake = field(converter=Snowflake)
    """The ID of the channel shown."""
    description: str = field()
    """A description shown with the channel."""
    emoji_id: str | Snowflake | None = field(converter=Snowflake, default=None)
    """The ID of the emoji if it isn't a unicode."""
    emoji_name: str | None = field(default=None)
    """The name of the emoji, if present."""

    @property
    def emoji(self) -> Emoji:
        return Emoji(id=self.emoji_id, name=self.emoji_name)


@define(kw_only=True)
class WelcomeScreen:
    """
    Represents the welcome screen in a guild from Discord.

    Attributes
    ----------
    description : `str`
        The description of the guild in the welcome screen.
    welcome_channels : `list[WelcomeScreenChannel]`, optional
        The channels show in the welcome screen. A maximum
        of `5` are able to be shown.

    Methods
    -------
    channels : `list[WelcomeScreenChannel]`, optional
        The channels show in the welcome screen. A maximum
        of `5` are shown.
    """

    description: str | None = field(default=None)
    """The description of the guild in the welcome screen."""
    welcome_channels: list[dict] | list[WelcomeScreenChannel] | None = field(
        converter=optional_c(list_c(WelcomeScreenChannel)), default=None
    )
    """
    The channels show in the welcome screen. A maximum
    of `5` are able to be shown.
    """

    @property
    def channels(self) -> list[WelcomeScreenChannel] | None:
        """
        The channels show in the welcome screen. A maximum
        of `5` are shown.
        """
        return self.welcome_channels


class SystemChannelFlags(IntFlag):
    """
    System channel flags are a set of bitwise values that represent
    the flags of a guild's welcome channel.
    """

    SUPPRESS_JOIN_NOTIFICATIONS = 1 << 0
    """Suppresses welcome messages from the channel."""
    SUPPRESS_PREMIUM_SUBSCRIPTIONS = 1 << 1
    """Suppresses server boosting messages from the channel."""
    SUPPRESS_GUILD_REMINDER_NOTIFICATIONS = 1 << 2
    """Suppresses server setup tips from the channel."""
    SUPPRESS_JOIN_NOTIFICATION_REPLIES = 1 << 3
    """Suppresses sticker reply pop-outs from the channel."""


class GuildNSFWLevel(IntEnum):
    """
    Represents the levels of NSFW filtering in a guild from Discord.

    Constants
    ---------
    DEFAULT
        There is a some filtering going on.
    EXPLICIT
        The guild is being checked for explicit NSFW content.
    SAFE
        The guild is not being checked for NSFW.
    AGE_RESTRICTED
        The guild is checked on parallel to `EXPLICIT` with
        verification required.
    """

    DEFAULT = 0
    """There is a some filtering going on."""
    EXPLICIT = 1
    """The guild is being checked for explicit NSFW content."""
    SAFE = 2
    """The guild is not being checked for NSFW."""
    AGE_RESTRICTED = 3
    """
    The guild is checked on parallel to `EXPLICIT` with
    verification required.
    """


class ExplicitContentFilterLevel(IntEnum):
    """
    Represents the explicit content filter levels of a guild
    from Discord.

    Constants
    ---------
    DISABLED
        Content is not being checked for explicit content.
    MEMBERS_WITHOUT_ROLES
        Content is being checked only from guild members
        without roles.
    ALL_MEMBERS
        Content is being checked from evey guild member.
    """

    DISABLED = 0
    """Content is not being checked for explicit content."""
    MEMBERS_WITHOUT_ROLES = 1
    """Content is being checked only from guild members without roles."""
    ALL_MEMBERS = 2
    """Content is being checked from evey guild member."""


class VerificationLevel(IntEnum):
    """
    Represents the verification levels of a guild from Discord.

    Constants
    ---------
    NONE
        There is no verification standard set.
    LOW
        Guild members must have their e-mail verified.
    MEDIUM
        Guild members must have been registered on Discord for longer than 5 minutes.
    HIGH
        Guild members must be part of the guild for longer than 10 minutes.
    VERY_HIGH
        Guild members must have their phone number verified.
    """

    NONE = 0
    """There is no verification standard set."""
    LOW = 1
    """Guild members must have their e-mail verified."""
    MEDIUM = 2
    """Guild members must have been registered on Discord for longer than 5 minutes."""
    HIGH = 3
    """Guild members must be part of the guild for longer than 10 minutes."""
    VERY_HIGH = 4
    """Guild members must have their phone number verified."""


@define(repr=False)
class UnavailableGuild(Partial, Object):
    """
    Represents an unavailable guild from Discord.

    Attributes
    ----------
    id : `Snowflake`
        The ID of the unavailable guild.
    unavailable : `bool`
        Whether the guild is unavailable or not.
        This is always set to `True` unless Discord
        says otherwise from a given payload. Use
        the class representation itself for this.
    """

    id: str | Snowflake = field(converter=Snowflake)
    unavailable: bool = field(default=True)

    def __repr__(self) -> bool:
        return self.unavailable


@define(kw_only=True)
class Guild(Object):
    """
    Represents a guild from Discord.


    Attributes
    ----------
    id : `Snowflake`
        The ID of the guild.
    name : `str`
        The name of the guild.
    icon : `str`
        The icon of the guild in a URL format.
    owner_id : `Snowflake`
        The ID of the owner of the guild.
    afk_timeout : `int`
        The current set AFK timeout in seconds for the guild.
    verification_level : `VerificationLevel`
        The set verification level for members of the guild.
    default_message_notifications : `int`
        The default notifications level of the guild.
    explicit_content_filter : `ExplicitContentFilterLevel`
        The explicit content filter level of the guild.
    features : `list[str]`
        The currently enabled features of the guild.
    mfa_level : `int`
        The required MFA (rep. by 2FA) level of the guild.
    system_channel_flags : `SystemChannelFlags`
        The guild's welcome channel's flags for suppression.
    premium_tier : `int`
        The current server boosting tier level of the guild.
    preferred_locale : `str`
        The preferred locale of the guild.
    nsfw_level : `int`
        The currently set NSFW level of the guild.
    premium_progress_bar_enabled : `bool`
        Whether the guild has the server boosting bar enabled or not.
    owner : `bool`
        Whether the user who invoked the guild is the owner or not.
    afk_channel_id : `Snowflake`, optional
        The ID of the AFK channel inside the guild, if present.
    icon_hash : `str`, optional
        The icon of the guild in a hash format.

        This hash is pre-determined by the API and does not reflect
        the URL path.
    splash : `str`, optional
        The hash of the guild's splash invite screen image, if present.
    discovery_splash : `str`, optional
        The hash of the guild's discovery splash screen image, if present.
    permissions : `str`, optional
        The calculated permissions of the user invoking the guild, if present.
    region : `str`, optional
        The ID of the voice region for the guild.

        This field has been deprecated as of `v8` and should no longer
        be used.
    widget_enabled : `bool`
        Whether the server has its widget enabled or not.
    widget_channel_id : `Snowflake`, optional
        The ID of the channel which the widget targets, if present.
    emojis : `list[Emoji]`, optional
        The Emojis that the guild owns.
    application_id : `Snowflake`, optional
        The ID of the application for the guild if created via. a bot.
    system_channel_id : `Snowflake`, optional
        The ID of the system welcome messages channel, if present.
    rules_channel_id : `Snowflake`, optional
        The ID of the rules channel, if presently determined as a Community server.
    max_presences : `int`, optional
        The maximum amount of presences allowed in the guild. Always set to `0`
        underneath a guild size cap.
    max_members : `int`, optional
        The maximum amount of members allowed in the guild.
        Globally set to `800000` currently.
    vanity_url_code : `str`, optional
        The vanity URL of the guild, if present.
    description : `str`, optional
        The description of the guild, if presently determined as a Community server.
    banner : `str`, optional
        The banner of the guild, if present.
    premium_subscription_count : `int`, optional
        The approximated count of boosts the guild has.
    public_updates_channel_id : `Snowflake`, optional
        The community moderation-only ID of the channel in the guild, if present.
    max_video_channel_users : `int`
        The maximum amount of users in a voice channel allowed to have video on.
        Globally set to `25` currently.
    approximate_member_count : `int`, optional
        The approximated member count of the guild.
    approximate_presence_count : `int`, optional
        The approximated amount of presences in the guild.
    welcome_screen : `WelcomeScreen`, optional
        The welcome screen of the guild, if present.
    stickers : `list[Sticker]`, optional
        The stickers that the guild owns.
    """

    id: str | Snowflake = field(converter=Snowflake)
    """The ID of the guild."""
    name: str = field()
    """The name of the guild."""
    icon: str = field()
    """The icon of the guild in a URL format."""
    owner_id: str | Snowflake = field(converter=Snowflake)
    """The ID of the owner of the guild."""
    afk_timeout: int = field()
    """The current set AFK timeout in seconds for the guild."""
    verification_level: int | VerificationLevel = field(converter=VerificationLevel)
    """The set verification level for members of the guild."""
    default_message_notifications: int = field()
    """The default notifications level of the guild."""
    explicit_content_filter: int | ExplicitContentFilterLevel = field(
        converter=ExplicitContentFilterLevel
    )
    """The explicit content filter level of the guild."""
    features: list[str] = field()
    """The currently enabled features inside of the guild."""
    mfa_level: int = field()
    """The required MFA (rep. by 2FA) level of the guild."""
    system_channel_flags: int | SystemChannelFlags = field(converter=SystemChannelFlags)
    """The guild's welcome channel's flags for suppression."""
    premium_tier: int = field()
    """The current server boosting tier level of the guild."""
    preferred_locale: str = field()
    """The preferred locale of the guild."""
    nsfw_level: int = field()
    """The currently set NSFW level of the guild."""
    premium_progress_bar_enabled: bool = field()
    """Whether the guild has the server boosting bar enabled or not."""
    owner: bool = field(default=False)
    """Whether the user who invoked the guild is the owner or not."""
    afk_channel_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """The ID of the AFK channel inside the guild, if present."""
    icon_hash: str | None = field(default=None)
    """
    The icon of the guild in a hash format.

    This hash is pre-determined by the API and does not reflect
    the URL path.
    """
    splash: str | None = field(default=None)
    """The hash of the guild's splash invite screen image, if present."""
    discovery_splash: str | None = field(default=None)
    """The hash of the guild's discovery splash screen image, if present."""
    permissions: str | None = field(default=None)
    """The calculated permissions of the user invoking the guild, if present."""
    region: str | None = field(default=None)
    """
    The ID of the voice region for the guild.

    This field has been deprecated as of `v8` and should no longer
    be used.
    """
    widget_enabled: bool = field(default=False)
    """Whether the server has its widget enabled or not."""
    widget_channel_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """The ID of the channel which the widget targets, if present."""
    roles: list[dict] | list[Role] | None = field(converter=optional_c(list_c(Role)), default=None)
    """The roles that the guild has, if present."""
    emojis: list[dict] | list[Emoji] | None = field(
        converter=optional_c(list_c(Emoji)), default=None
    )
    """The Emojis that the guild owns."""
    application_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """The ID of the application for the guild if created via. a bot."""
    system_channel_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """The ID of the system welcome messages channel, if present."""
    rules_channel_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """The ID of the rules channel, if presently determined as a Community server."""
    max_presences: int | None = field(default=None)
    """
    The maximum amount of presences allowed in the guild. Always set to `0`
    underneath a guild size cap."""
    max_members: int | None = field(default=None)
    """The maximum amount of members allowed in the guild.
    Globally set to `800000` currently.
    """
    vanity_url_code: str | None = field(default=None)
    """The vanity URL of the guild, if present."""
    description: str | None = field(default=None)
    """The description of the guild, if presently determined as a Community server."""
    banner: str | None = field(default=None)
    """The banner of the guild, if present."""
    premium_subscription_count: int | None = field(default=None)
    """The approximated count of boosts the guild has."""
    public_updates_channel_id: str | Snowflake | None = field(
        converter=optional_c(Snowflake), default=None
    )
    """The community moderation-only ID of the channel in the guild, if present."""
    max_video_channel_users: int = field(default=25)
    """
    The maximum amount of users in a voice channel allowed to have video on.
    Globally set to `25` currently.
    """
    approximate_member_count: int | None = field(default=None)
    """The approximated member count of the guild."""
    approximate_presence_count: int | None = field(default=None)
    """The approximated amount of presences in the guild."""
    welcome_screen: dict | WelcomeScreen | None = field(
        converter=optional_c(WelcomeScreen), default=None
    )
    """The welcome screen of the guild, if present."""
    stickers: list[dict] | list[Sticker] | None = field(converter=Sticker, default=None)
    """The stickers that the guild owns."""


@define(kw_only=True)
class GuildPreview(Object):
    """
    Represents the preview of a guild from Discord.

    Attributes
    ----------
    id : `Snowflake`
        The ID of the guild being previewed.
    name : `str`
        The name of the guild being previewed.
    features : `list[str]`
        The enabled features of the previewed guild.
    emojis : `list[Emoji]`
        The guild's custom Emojis.
    approximate_member_count : `int`
        The approximated amount of members in the previewed guild.
    approximate_presence_count : `int`
        The approximated count of presences in the previewed guild.
    icon : `str`, optional
        The icon of the guild being previewed, if present.
    splash : `str,` optional
        The splash invite background screen image of the previewed
        guild, if present.
    discovery_splash : `str`, optional
        The hash discovery splash screen image, of the previewed
        guild, if present.
    description : `str`, optional
        The description of the guild being previewed, if presently
        determined as a Community server.
    stickers : `list[Sticker]`, optional
        The stickers of the guild being previewed.
    """

    id: str | Snowflake = field(converter=Snowflake)
    """The ID of the guild being previewed."""
    name: str = field()
    """The name of the guild being previewed."""
    features: list[str] = field()
    """The enabled features of the previewed guild."""
    emojis: list[dict] | list[Emoji] = field(converter=list_c(Emoji))
    """The guild's custom Emojis."""
    approximate_member_count: int = field()
    """The approximated amount of members in the previewed guild."""
    approximate_presence_count: int = field()
    """The approximated count of presences in the previewed guild."""
    icon: str | None = field(default=None)
    """The icon of the guild being previewed, if present."""
    splash: str | None = field(default=None)
    """
    The splash invite background screen image of the previewed
    guild, if present.
    """
    discovery_splash: str | None = field(default=None)
    """
    The hash discovery splash screen image, of the previewed
    guild, if present.
    """
    description: str | None = field(default=None)
    """
    The description of the guild being previewed, if presently
    determined as a Community server.
    """
    stickers: list[dict] | list[Sticker] | None = field(converter=optional_c(list_c(Sticker)), default=None)
    """The stickers of the guild being previewed."""


@define(kw_only=True)
class GuildWidget(Object):
    """Represents the widget of a guild from Discord."""

    id: str | Snowflake = field(converter=Snowflake)
    """The ID of the guild on the widget."""
    name: str = field()
    """The name of the guild on the widget."""
    # TODO: implement Partial Channel object.
    # channels: list[dict] | list[PartialChannel] = field(converter=list_c(PartialChannel))
    # TODO: implement Partial Member object.
    # members: list[dict] | list[PartialMember] = field(converter=list_c(PartialMember))
    presence_count: int = field()
    """The amount of members online in the guild."""
    instant_invite: str | None = field(default=None)
    """An instant invite provided for the guild if it exists."""


@define(kw_only=True)
class GuildWidgetSettings:
    """
    Represents the settings of a widget of a guild in Discord.

    Attributes
    ----------
    enabled : `bool`
        Whether the widget is enabled for the guild or not.
    channel_id : `Snowflake`, optional
        The ID of the channel shown in the guild's widget,
        if present.
    """

    enabled: bool = field()
    """Whether the widget is enabled for the guild or not."""
    channel_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """The ID of the channel shown in the guild's widget, if present."""


@define(kw_only=True)
class Member(Partial):
    """
    Represents the member in a guild from Discord.

    ---

    Guild members have their base user information
    present via. accessing the `user` field. This information
    is presented in the form of property methods, in order
    to avoid making users go through an unnecessary chain step.

    This dataclass is registered as a "partial" because it derives
    some information pertaining to a `User` object. Despite this,
    it also registers as an object albeit the lack of the `id` field
    being properly given by the Gateway.

    ---

    Attributes
    ----------
    user : `User`, optional
        The user representation of the member in the guild. This is only
        `None` when provided in a `MESSAGE_CREATE` or `MESSAGE_UPDATE`
        Gateway event.
    nick : `str`, optional
        The nickname of the member in the guild, if present.
    avatar : `str`, optional
        The hash of the guild member's avatar, if present.
    roles : `list[Snowflake]`, optional
        The roles of the member in the guild, if present.

        Roles are only provided in the form of `Snowflake` objects,
        and are not given as a `Role` object. This is subject to change
        in the future.
    joined_at : `datetime`
        The time at which the member joined the guild.
    premium_since : `datetime`, optional
        The time at which the member began boosting the guild,
        if present.
    deaf : `bool`
        Whether the member is deafened in any of the guild's voice channels
        or not. Defaults to `False`.
    mute : `bool`
        Whether the member is muted in any of the guild's voice channels
        or not. Defaults to `False`.
    pending : `bool`
        Whether the member is still pending access to join the guild
        or not. Defaults to `False`. This is only supplied when any Gateway
        event that is not intended for guilds.
    permissions : `str`, optional
        The calculated permissions of the member in the guild, including
        any overwrites. This is only returned when passed inside an
        `Interaction` object.
    communication_disabled_until : `datetime`, optional
        The time remaining until the member of the guild has their
        timeout removed. This is `None` when a timeout has not yet
        been applied, or when one has ended.

    Methods
    -------
    id : `Snowflake`, optional
        The ID of the user, if present.
    username : `str,` optional
        The ID of the user, if present.
    discriminator : `str`, optional
        The discriminator (4-digit tag) of the user, if present.
    avatar : `str`, optional
        The hash of the user's avatar, if present.
    bot : `bool`, optional
        Whether the user is a bot or not, if present.
    system : `bool`, optional
        Whether the user is from the official Discord System or not,
        if present.
    mfa_enabled : `bool`, optional
        Whether the user has 2FA (two-factor authentication) enabled
        or not, if present.
    banner : `str`, optional
        The hash of the user's banner, if present.
    accent_color : `int`, optional
        The color of the user's banner, if present.
    locale : `str`, optional
        The user's selected locale, if present.
    verified : `bool`, optional
        Whether the user has a verified e-mail or not.
    email : `str`, optional
        The e-mail associated to the user's account, if present.
    flags : `UserFlags`, optional
        The public flags on the user's account, if present.
    premium_type : `UserPremiumType`, optional
        The type of Nitro subscription the user has, if present.
    public_flags : `UserFlags`, optional
        The type of Nitro subscription the user has, if present.
    """

    user: dict | User | None = field(converter=optional_c(User), default=None)
    """
    The user representation of the member in the guild. This is only `None` when
    provided in a `MESSAGE_CREATE` or `MESSAGE_UPDATE` Gateway event.
    """
    nick: str | None = field(default=None)
    """The nickname of the member in the guild, if present."""
    avatar: str | None = field(default=None)
    """The hash of the guild member's avatar, if present."""
    # TODO: change roles field to give a list of Role objects. This is an API
    # limitation.
    roles: list[str] | list[Snowflake] | None = field(
        converter=optional_c(list_c(Snowflake)), default=None
    )
    """
    The roles of the member in the guild, if present.

    Roles are only provided in the form of `Snowflake` objects, and are not given
    as a `Role` object. This is subject to change in the future.
    """
    joined_at: str | datetime = field(converter=datetime.fromisoformat)
    """The time at which the member joined the guild."""
    premium_since: str | datetime | None = field(
        converter=optional_c(datetime.fromisoformat), default=None
    )
    """The time at which the member began boosting the guild, if present."""
    deaf: bool = field(default=False)
    """
    Whether the member is deafened in any of the guild's voice channels or not.
    Defaults to `False`.
    """
    mute: bool = field(default=False)
    """
    Whether the member is muted in any of the guild's voice channels or not.
    Defaults to `False`.
    """
    pending: bool = field(default=False)
    """
    Whether the member is still pending access to join the guild or not.
    Defaults to `False`. This is only supplied when any Gateway event that
    is not intended for guilds.
    """
    permissions: str | None = field(default=None)
    """
    The calculated permissions of the member in the guild, including
    any overwrites. This is only returned when passed inside of an
    `Interaction` object.
    """
    communication_disabled_until: str | datetime | None = field(
        converter=optional_c(datetime.fromisoformat), default=None
    )
    """
    The time remaining until the member of the guild has their timeout
    removed. This is `None` when a timeout has not yet  been applied,
    or when one has ended.
    """

    @staticmethod
    def exists(base: Any, attr: Any) -> Any | None:
        """
        Determines if a field exists for property methods.

        Attributes
        ----------
        base : `typing.Any`
            The field to check for. This is usually a class
            given.
        attr : `typing.Any`
            The specific attribute of the base to look for.
            This works off of `base`, which should be an existing
            attribute of what you're looking at inside the given
            dataclass.

        Returns
        -------
        `typing.Any`, optional
            The value of the attribute, if it is present.
            This will otherwise return `None`.
        """
        if base is not None:
            return attr
        else:
            return None

    @property
    def id(self) -> Snowflake | None:
        """The ID of the user, if present."""
        return self.exists(self.user, self.user.id)

    @property
    def username(self) -> str | None:
        """The ID of the user, if present."""
        return self.exists(self.user, self.user.username)

    @property
    def discriminator(self) -> str | None:
        """The discriminator (4-digit tag) of the user, if present."""
        return self.exists(self.user, self.user.discriminator)

    @property
    def avatar(self) -> str | None:
        """The hash of the user's avatar, if present."""
        if _avatar := self.exists(self.user, self.user.avatar):
            return _avatar
        return self.avatar

    @property
    def bot(self) -> bool | None:
        """Whether the user is a bot or not, if present."""
        return self.exists(self.user, self.user.bot)

    @property
    def system(self) -> bool | None:
        """Whether the user is from the official Discord System or not, if present."""
        return self.exists(self.user, self.user.system)

    @property
    def mfa_enabled(self) -> bool | None:
        """
        Whether the user has 2FA (two-factor authentication) enabled
        or not, if present.
        """
        return self.exists(self.user, self.user.mfa_enabled)

    @property
    def banner(self) -> str | None:
        """The hash of the user's banner, if present."""
        return self.exists(self.user, self.user.banner)

    @property
    def accent_color(self) -> int | None:
        """The color of the user's banner, if present."""
        return self.exists(self.user, self.user.accent_color)

    @property
    def locale(self) -> str | None:
        """The user's selected locale, if present."""
        return self.exists(self.user, self.user.locale)

    @property
    def verified(self) -> bool | None:
        """Whether the user has a verified e-mail or not."""
        return self.exists(self.user, self.user.verified)

    @property
    def email(self) -> str | None:
        """The e-mail associated to the user's account, if present."""
        return self.exists(self.user, self.user.email)

    @property
    def flags(self) -> UserFlags | None:
        """The public flags on the user's account, if present."""
        return self.exists(self.user, self.user.flags)

    @property
    def premium_type(self) -> UserPremiumType | None:
        """The type of Nitro subscription the user has, if present."""
        return self.exists(self.user, self.user.premium_type)

    @property
    def public_flags(self) -> UserFlags | None:
        """The type of Nitro subscription the user has, if present."""
        return self.exists(self.user, self.user.public_flags)



from enum import IntEnum

from attrs import define, field

from ...utils.converters import list_c

from .abc import Component

__all__ = (
    "ActionRow",
    "Button",
    "SelectMenu",
    "SelectOption",
    "Modal",
    "TextInput",
    "ComponentType",
    "ButtonStyle",
    "TextInputStyle",
)


class ComponentType(IntEnum):
    """
    Represents the types of components from Discord.

    Constants
    ---------
    ACTION_ROW
        An action row, used in `ActionRow`.
    BUTTON
        A button, used in `Button`.
    SELECT_MENU
        A select menu, used in `SelectMenu`.
    TEXT_INPUT
        A text input, used in `TextInput`.
    """

    ACTION_ROW = 1
    """An action row, used in `ActionRow`."""
    BUTTON = 2
    """A button, used in `Button`."""
    SELECT_MENU = 3
    """A select menu, used in `SelectMenu`."""
    TEXT_INPUT = 4
    """A text input, used in `TextInput`."""


@define(kw_only=True)
class ActionRow(Component):
    """
    Represents an action row component from Discord.

    ---

    An action row may only have 5 components stored in it.
    It may not be nested inside of itself or
    contain itself.

    ---

    Attributes
    ----------
    type : `ComponentType`
        The type of component, as `ComponentType.ACTION_ROW`.
    components : `list[Component]`
        The components stored under the action row.
    """

    type: int | ComponentType = field(converter=ComponentType, default=ComponentType.ACTION_ROW)
    """The type of component, as `ComponentType.ACTION_ROW`."""
    components: list[Component] = field(converter=list_c(Component))
    """The components stored in the action row."""


class ButtonStyle(IntEnum):
    """
    Represents the styles of buttons from Discord.

    ---

    See `Button` for information regarding the use of styles.

    ---

    Constants
    ---------
    PRIMARY (BLUE)
        A "CTA" indication.
    SECONDARY (GREY)
        A secondary usage indication.
    SUCCESS (GREEN)
        A successful indication.
    DANGER (RED)
        A dangerous or destructive indication.
    LINK (URL)
        An navigation indication.
    """

    PRIMARY = 1
    """A \"CTA\" indication."""
    SECONDARY = 2
    """A secondary usage indication."""
    SUCCESS = 3
    """A successful indication."""
    DANGER = 4
    """A dangerous or destructive indication."""
    LINK = 5
    """An navigation indication."""

    # These are just aliases of the button styles.
    BLUE = 1
    """An alias of `PRIMARY`."""
    GREY = 2
    """An alias of `SECONDARY`."""
    GREEN = 3
    """An alias of `SUCCESS`."""
    RED = 4
    """An alias of `DANGER`."""
    URL = 5
    """An alias of `LINK`."""


@define(kw_only=True)
class Button(Component):
    """
    Represents a button component from Discord.

    ---

    To use a style that indicates anything besides a navigation path,
    (or essentially a colour) you must have the `custom_id` attribute
    filled in your component. The `url` attribute is not allowed.

    For a link, the `url` attribute must be present without `custom_id`.

    ---

    Attributes
    ----------
    type : `ComponentType`
        The type of component, as `ComponentType.BUTTON`.
    style : `ButtonStyle`
        The style of the button. Please see `Button` for usage.
    label : `str`, optional
        The labelled contents of the button. The maximum
        length is 80 characters.
    custom_id : `str`, optional
        A customisable identifiable descriptor for the button.
    url : `str`, optional
        The URL relating to the button.
    disabled : `bool`
        Whether the button is disabled for usage or not. Defaults to `False`.
    """

    type: int | ComponentType = field(converter=ComponentType, default=ComponentType.BUTTON)
    """The type of component, as `ComponentType.BUTTON`."""
    style: int | ButtonStyle = field(converter=ButtonStyle)
    """The style of the button. Please see `Button` for usage."""
    label: str | None = field(default=None)
    """The labelled contents of the button. The maximum length is 80 characters."""

    # TODO: implement a partial emoji abc.
    # emoji: dict | PartialEmoji | None = field(converter=PartialEmoji, default=None)
    # """The emoji related to the button, if present."""

    custom_id: str | None = field(default=None)
    """A customisable identifiable descriptor for the button."""
    url: str | None = field(default=None)
    """The URL relating to the button."""
    disabled: bool = field(default=False)
    """Whether the button is disabled for usage or not. Defaults to `False`."""


@define(kw_only=True)
class SelectOption:
    """
    Represents an option in a select menu from Discord.

    Attributes
    ----------
    label : `str`
        The labelled contents of the select option.
        The maximum length is 80 characters.
    value : `str`
        The value of the select option.
    description : `str`, optional
        An additional description of the select option, if present.
    default : `bool`
        Whether the select option is chosen by default or not.
        Defaults to `False`.
    """

    label: str = field()
    """The label of the select option."""
    value: str = field()
    """The value of the select option."""
    description: str | None = field(default=None)
    """An additional description of the select option, if present."""

    # TODO: implement a partial emoji abc.
    # emoji: dict | PartialEmoji | None = field(default=None)
    # """The emoji related to the select option, if present."""

    default: bool = field(default=False)
    """Whether the select option is chosen by default or not. Defaults to `False`."""


@define(kw_only=True)
class SelectMenu(Component):
    """
    Represents a select menu component from Discord.

    Attributes
    ----------
    type : `ComponentType`
        The type of component, as `ComponentType.SELECT_MENU`.
    custom_id : `str`
        A customisable identifiable descriptor for the select menu.
    options : `list[SelectOption]`
        The options of the select menu.
    placeholder : `str`, optional
        The placeholder text of the select menu, if present.
    min_values : `int`
        The minimum number of options needed for the menu in-between 0-25.
        Defaults to `1`.
    max_values : `int`
        The maximum number of options needed for the menu in-between 1-25.
        Defaults to `1`.
    disabled : `bool`
        Whether the select menu is disabled for usage or not. Defaults
        to `False`.
    """

    type: int | ComponentType = field(converter=ComponentType, default=ComponentType.SELECT_MENU)
    """The type of component, as `ComponentType.SELECT_MENU`."""
    custom_id: str = field()
    """A customisable identifiable descriptor for the select menu."""
    options: list[SelectOption] = field(converter=list_c(SelectOption))
    """The options of the select menu."""
    placeholder: str | None = field(default=None)
    """The placeholder text of the select menu, if present."""
    min_values: int = field(default=1)
    """
    The minimum number of options needed for the menu
    in-between 0-25. Defaults to `1`.
    """
    max_values: int = field(default=1)
    """
    The maximum number of options needed for the menu
    in-between 1-25. Defaults to `1`.
    """
    disabled: bool = field(default=False)
    """Whether the select menu is disabled for usage or not. Defaults to `False`."""


class TextInputStyle(IntEnum):
    """
    Represents the styles of text inputs from Discord.

    Constants
    ---------
    SHORT
        A short-form/single-line input.
    PARAGRAPH
        A long-form/multi-line input.
    """

    SHORT = 1
    """A short-form/single-line input."""
    PARAGRAPH = 2
    """A long-form/multi-line input."""


@define(kw_only=True)
class TextInput(Component):
    """
    Represents a text input component from Discord.

    Attributes
    ----------
    type : `ComponentType`
        The type of component, as `ComponentType.TEXT_INPUT`.
    custom_id : `str`
        A customisable identifiable descriptor for the text input.
    style : `TextInputStyle`
        The style of the text input.
    label : `str`
        The labelled contents of the modal. The maximum length is 45 characters.
    min_length : `int`, optional
        The minimum allowed length for a text input in-between
        1-4000 characters. Defaults to `0`.
    max_length : `int`, optional
        The maximum allowed length for a text input in-between
        1-4000 characters. Defaults to `1`.
    required : `bool`, optional
        Whether the text input is required to be filled or not. Defaults to `True`.
    value : `str`, optional
        A pre-filled value for the text input, if present.
        The maximum length is 4000 characters.
    placeholder : `str`, optional
        The placeholder text of the text input, if present.
        The maximum length is 100 characters.
    """

    type: int | ComponentType = field(converter=ComponentType, default=ComponentType.TEXT_INPUT)
    """The type of component, as `ComponentType.TEXT_INPUT`."""
    custom_id: str = field()
    """A customisable identifiable descriptor for the text input."""
    style: int | TextInputStyle = field(converter=TextInputStyle)
    """The style of the text input."""
    label: str = field()
    """The labelled contents of the modal. The maximum length is 45 characters."""
    min_length: int = field(default=0)
    """
    The minimum allowed length for a text input in-between
    1-4000 characters. Defaults to `0`.
    """
    max_length: int = field(default=1)
    """
    The maximum allowed length for a text input in-between
    1-4000 characters. Defaults to `1`.
    """
    required: bool = field(default=True)
    """Whether the text input is required to be filled or not. Defaults to `True`."""
    value: str | None = field(default=None)
    """A pre-filled value for the text input, if present. The maximum length is 4000 characters."""
    placeholder: str | None = field(default=None)
    """The placeholder text of the text input, if present. The maximum length is 100 characters."""


@define(kw_only=True)
class Modal(Component):
    """
    Represents a modal component from Discord.

    Attributes
    ----------
    title : `str`
        The title of the modal.
    custom_id : `str`
        A customisable identifiable descriptor for the modal.
    components : `list[TextInput]`
        The components stored in the modal.
    """

    title: str = field()
    """The title of the modal."""
    custom_id: str = field()
    """A customisable identifiable descriptor for the modal."""
    components: list[TextInput] = field(converter=list_c(TextInput))
    """The components stored in the modal."""



from attrs import define, field

from ...client.resources.guild import UnavailableGuild
from ...client.resources.application import PartialApplication
from ...utils import list_c

from .abc import _Event

__all__ = ("Ready", "HeartbeatAck", "Resumed", "Reconnect", "InvalidSession")


@define()
class Ready(_Event):
    """
    Represents when the client has successfully connected.

    ---

    The `READY` event can be the largest and most complex one
    the Gateway will send, as it contains all the state required
    for a client to begin interacting with the rest of the platform.

    ---

    Attributes
    ----------
    v : `int`
        The used version of the Discord API.
    user_settings : `dict`, optional
        The settings of the bot application, if present.
    guilds : `list[UnavailableGuild]`
        The guilds unavailable to the bot.
    geo_ordered_rtc_regions : `list[str]`
        The RTC voice regions accessible to the bot application.
    session_type : `str`
        The type of session that the bot has established with the Gateway.
    session_id : `str`
        The ID of the bot's Gateway connection session, used for reconnection.
    resume_gateway_url : `str`
        The URL of the Gateway upon resuming an existing connection.
    shard : `list[int]`, optional
        The shards of the Gateway connection, if present.
    application : `PartialApplication`
        The application form of the bot. Contains only `id` and `flags`.

    Methods
    -------
    version : `int`
        The used version of the Discord API.
    """

    v: int = field(kw_only=True)
    """The used version of the Discord API."""
    user_settings: dict | None = field(default=None, kw_only=True)
    """The settings of the bot application, if present."""
    # TODO: implement User object
    user: dict = field(kw_only=True)
    """The user form of the bot application."""
    guilds: list[dict] | list[UnavailableGuild] = field(
        converter=list_c(UnavailableGuild), kw_only=True
    )
    """The guilds unavailable to the bot."""
    # TODO: Investigate the guild_join_requests field.
    guild_join_requests: list | None = field(default=None, kw_only=True)
    """
    The pending approval requests for guilds the bot cannot access.

    Some data is conferred here supplied from `guilds`.
    """
    geo_ordered_rtc_regions: list[str] = field(kw_only=True)
    """The RTC voice regions accessible to the bot application."""
    session_type: str = field(kw_only=True)
    """The type of session that the bot has established with the Gateway."""
    session_id: str = field(kw_only=True)
    """The ID of the bot's Gateway connection session, used for reconnection."""
    resume_gateway_url: str = field(kw_only=True)
    """The URL of the Gateway upon resuming an existing connection."""
    # TODO: Investigate the relationships field.
    relationships: list | None = field(default=None, kw_only=True)
    """The relationships associated to the bot application, if present."""
    # TODO: Investigate the private_channels field.
    private_channels: list | None = field(default=None, kw_only=True)
    """The private channels of the bot application, if present."""
    # TODO: Investigate the presences field.
    presences: list | None = field(default=None, kw_only=True)
    """The presences of the bot application, if present."""
    shard: list[int] | None = field(default=None, kw_only=True)
    """The shards of the Gateway connection, if present."""
    application: dict | PartialApplication = field(converter=PartialApplication, kw_only=True)
    """The application form of the bot. Contains only `id` and `flags`."""

    @property
    def version(self) -> int:
        """The used version of the Discord API."""
        return self.v


@define()
class HeartbeatAck(_Event):
    """
    Represents when the client's Gateway connection has validated a heartbeat.

    Attributes
    ----------
    latency : `float`
        The latency or difference in milliseconds between heartbeats.
    """

    latency: float = field(kw_only=True)
    """The latency or difference in milliseconds between heartbeats."""


@define()
class Resumed(_Event):
    """
    Represents when the client has successfully resumed a connection.

    Attributes
    ----------
    token : `str`
        The token of the bot used.
    session_id : `str`
        The ID of the Gateway connection session.
    seq : `int`
        The last sequence number given for the session.
    """

    token: str = field(kw_only=True)
    """The token of the bot used."""
    session_id: str = field(kw_only=True)
    """The ID of the Gateway connection session."""
    seq: int = field(kw_only=True)
    """The last sequence number given for the session."""


@define()
class Reconnect(_Event):
    """
    Represents when the client has been told to reconnect.

    ---

    It should be noted that reconnection is automatically handled for you
    in retux already. This event should never be used outside of logging down
    when reconnections were made. Do not use this event to force a reconnection
    unless you know what you're doing!
    """


@define(repr=False)
class InvalidSession(_Event):
    """
    Represents when the client has an invalidated Gateway connection.

    ---

    Invalid sessions are determined by one of three factors:

    - The Gateway could not resume the connection.
    - The client could not identify itself correctly.
    - The Gateway has invalidated the session for you, and requires a new connection.

    ---

    Attributes
    ----------
    _invalid_session : `bool`
        Whether the session can be reconnected to or not.
        This should never need to be called upon directly.
        Please use the representation of the class itself.

        This defaults to `False` and determined on good faith
        of ensuring a stable connection over potential client
        conflict.

    Methods
    -------
    can_reconnect() : `bool`
        Checks the session to see if a reconnection is possible.
    """

    _invalid_session: bool = field(default=False)
    """
    Whether the session can be reconnected to or not.
    This should never need to be called upon directly.
    Please use the representation of the class itself.

    This defaults to `False` and determined on good faith
    of ensuring a stable connection over potential client
    conflict.
    """

    def __repr__(self) -> bool:
        return self._invalid_session

    def can_reconnect(self) -> bool:
        """
        Checks the session to see if a reconnection is possible.

        Returns
        -------
        `bool`
            Whether the session can be reconnected to or not.
        """
        return self._invalid_session



class HTTPException(Exception):
    """An error occurred trying to run an HTTP request through the REST API."""

    payload: dict
    """The error trace stack of the exception."""
    message: str
    """The message body of the exception."""
    code: int
    """The error code to correlate the exception to."""
    severity: int
    """The severity level to correlate the exception to. This is treated for logging purposes."""

    def __init__(self, payload: dict, severity: int = 0):
        """
        Creates a new exception for HTTP requests.

        Parameters
        ----------
        payload : `str`
            The error trace stack of the exception.
        severity : `int`, optional
            The severity level to correlate the exception to. This is treated for logging purposes.
        """
        self.payload = payload
        self.message = self.payload.get("message")
        self.code = self.payload.get("code")
        self.severity = severity

        self._parse()

    def _parse(self):
        """Parses the contents of the error given and prepares for printout."""

        # TODO: Figure out a way to make the error payload appear as a -> tree from
        # the dictionary keys, and exclude _errors.
        # ex: activities -> 0 -> platform -> Value must be one of ('desktop', 'android', 'ios'). (err. 50035, Invalid Form Body)

        path_to_err = f"{self.payload.get('code')} -> {self.payload.get('message')}"
        self._printout(path_to_err)

    def _printout(self, content: str):
        """Prints out the content given and raises as an exception.

        Parameters
        ----------
        content : `str`
            The content of the exception.

        Raises
        ------
        HTTPException
        """
        super().__init__(content)


class InvalidToken(Exception):
    """An invalid token was supplied to the Gateway."""


class RateLimited(Exception):
    """
    Too many Gateway commands were sent while being connected.

    You have been disconnected from the Gateway as a result.
    """


class InvalidShard(Exception):
    """An invalid shard was supplied to the Gateway."""


class RequiresSharding(Exception):
    """
    The connection in particular requires sharding, as the guild volume is
    too large to process on one concurrent lay line.
    """


class InvalidIntents(Exception):
    """
    An invalid intent or series of intents were supplied to the Gateway.

    Please check to make sure you've calculated them correctly and are
    using the `|` join operator.
    """


class DisallowedIntents(Exception):
    """
    A disallowed intent or series of intents were supplied to the Gateway.

    You may only supply intents that you have been approved or allowed for.
    If your bot application is pending verification and/or is missing an intent
    checked off in the Developer Portal, this may be the reason why.
    """



from .abc import Snowflake, Object
from .user import User
from .guild import Member
from ...utils.converters import optional_c

from attrs import define, field
from datetime import datetime
from enum import IntEnum

__all__ = (
    "GuildScheduledEvent",
    "GuildScheduledEventEntityType",
    "GuildScheduledEventStatus",
    "GuildScheduledEventEntityMetadata",
    "GuildScheduledEventPrivacyLevel",
    "GuildScheduledEventUser",
)


class GuildScheduledEventPrivacyLevel(IntEnum):
    """
    Represents the privacy levels of a guild scheduled event from Discord.

    Constants
    ---------
    GUILD_ONLY
        The scheduled event is only accessible to guild members.
    """

    GUILD_ONLY = 2
    """The scheduled event is only accessible to guild members."""


class GuildScheduledEventEntityType(IntEnum):
    """
    Represents the different types of guild scheduled events from Discord.

    Constants
    ---------
    STAGE_INSTANCE
        The event will happen in a stage instance.
    VOICE
        The event will happen in a voice channel.
    EXTERNAL
        The event will happen in an external location.

        This requires `GuildScheduledEventEntityMetadata`.
    """

    STAGE_INSTANCE = 1
    """The event will happen in a stage instance."""
    VOICE = 2
    """The event will happen in a voice channel."""
    EXTERNAL = 3
    """The event will happen in an external location."""


class GuildScheduledEventStatus(IntEnum):
    """
    Represents the different status a guild scheduled event can have.

    Once the status is set to `Completed` or `Canceled`, it can't be updated again.

    Constants
    ----------
    SCHEDULED
        The event is currently scheduled and not active.
    ACTIVE
        The event is happening right now.
    COMPLETED
        The event was completed successfully.
    CANCELED
        The event was canceled before it could begin.
    """

    SCHEDULED = 1
    """The event is currently scheduled and not active."""
    ACTIVE = 2
    """The event is happening right now."""
    COMPLETED = 3
    """The event was completed successfully."""
    CANCELED = 4
    """The event was canceled before it could begin."""


@define(kw_only=True)
class GuildScheduledEventEntityMetadata:
    """
    Represents the additional metadata of guild scheduled events from Discord.

    Attributes
    ----------
    location : `str`, optional
        The location of the event in-between 1-100 characters.

        This is required for events with the `EXTERNAL` entity type.
    """

    location: str | None = field(default=None)
    """The location of the event in-between 1-100 characters."""


@define(kw_only=True)
class GuildScheduledEvent(Object):
    """
    Represents a scheduled event of a guild from Discord.

    Attributes
    ----------
    id : `Snowflake`
        The ID belonging to the scheduled event.
    guild_id : `Snowflake`
        The ID of the guild the event belongs to.
    channel_id : `Snowflake`, optional
        The ID of the channel the event will be hosted in, if any.

        This is not available for Events with the `entity_type` `EXTERNAL`.
    creator_id : `Snowflake`, optional
        The ID of the user who created the event, if any.

        This is only available for events created after October 25th, 2021.
    namer : `str`
        The name of the scheduled event (1-100 characters).
    description : `str`, optional
        The description of the scheduled event, if any (1-1000 characters).
    scheduled_start_time : `datetime`
        The time the scheduled event will start.
    scheduled_end_time : `datetime`, optional
        The time the scheduled event will end, if any.

        This is only required if the `entity_type` is `EXTERNAL`
    privacy_level : `GuildScheduledEventPrivacyLevel`
        The privacy level of the scheduled event.
    status : `GuildScheduledEventStatus`
        The status of the scheduled event.
    entity_type : `GuildScheduledEventEntityType`
        The type of the scheduled event.
    entity_id : `Snowflake`, optional
        The ID of an entity associated with a guild scheduled event, if any.
    entity_metadata : `GuildScheduledEventEntityMetadata`, optional
        Additional metadata for the guild scheduled event, if any.

        This must not present for scheduled events with the type `VOICE` or `STAGE_INSTANCE` but must be set for
        scheduled events with the type `EXTERNAL`.
    creator : `User`, optional
        The user that created the scheduled event.
    user_count : `int`, optional
        The number of users subscribed to the scheduled event.
    image : `str`, optional
        The cover image hash of the scheduled event.
    """

    id: str | Snowflake = field(converter=Snowflake)
    """The ID belonging to the scheduled event."""
    guild_id: str | Snowflake = field(converter=Snowflake)
    """The ID of the guild the event belongs to."""
    channel_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """The ID of the channel the event will be hosted in."""
    creator_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """The ID of the user who created the event."""
    name: str = field()
    """The name of the scheduled event (1-100 characters)."""
    description: str | None = field(default=None)
    """The description of the scheduled event (1-1000 characters)."""
    scheduled_start_time: str | datetime = field(converter=datetime.fromisoformat)
    """The time the scheduled event will start."""
    scheduled_end_time: str | datetime | None = field(
        converter=optional_c(datetime.fromisoformat), default=None
    )
    """The time the scheduled event will end."""
    privacy_level: int | GuildScheduledEventPrivacyLevel = field(
        converter=GuildScheduledEventPrivacyLevel
    )
    """The privacy level of the scheduled event."""
    status: int | GuildScheduledEventStatus = field(converter=GuildScheduledEventStatus)
    """The status of the scheduled event."""
    entity_type: int | GuildScheduledEventEntityType = field(
        converter=GuildScheduledEventEntityType
    )
    """The type of the scheduled event."""
    entity_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """The ID of an entity associated with a guild scheduled event."""
    entity_metadata: dict | GuildScheduledEventEntityMetadata | None = field(
        converter=optional_c(GuildScheduledEventEntityMetadata), default=None
    )
    """Additional metadata for the guild scheduled event."""
    creator: dict | User | None = field(converter=optional_c(User), default=None)
    """The user that created the scheduled event."""
    user_count: int | None = field(default=None)
    """The number of users subscribed to the scheduled event."""
    image: str | None = field(default=None)
    """The cover image hash of the scheduled event."""


class GuildScheduledEventUser:
    """
    Represents a user in a guild scheduled event from Discord.

    Attributes
    ----------
    guild_scheduled_event_id : `Snowflake`
        The scheduled event ID which the user subscribed to.
    user : `User`
        The user which subscribed to an event.
    member : `Member`, optional
        Guild member data for this user for the guild which this event belongs to, if any.
    """

    guild_scheduled_event_id: str | Snowflake = field(converter=Snowflake)
    """The scheduled event id which the user subscribed to."""
    user: dict | User = field(converter=User)
    """The user which subscribed to an event."""
    member: dict | Member | None = field(converter=optional_c(Member), default=None)
    """Guild member data for this user for the guild which this event belongs to, if any."""



from .abc import Snowflake, Object
from .user import User
from .guild import Guild
from .channel import Channel
from ...utils.converters import optional_c

from attrs import define, field
from enum import IntEnum


class WebhookType(IntEnum):
    """
    Represents the types of webhooks from Discord.

    Constants
    ---------
    INCOMING
        Incoming webhooks can post messages to channels with a generated token.
    CHANNEL_FOLLOWER
        Channel-Follower webhooks are internal webhooks used with channel following to post new messages into channels.
    APPLICATION
        Application webhooks are webhooks used with interactions.
    """

    INCOMING = 1
    """Incoming webhooks can post messages to channels with a generated token."""
    CHANNEL_FOLLOWER = 2
    """Channel-Follower webhooks are internal webhooks used with channel following to post new messages into channels."""
    APPLICATION = 3
    """Application webhooks are webhooks used with interactions."""


@define()
class Webhook(Object):
    """
    Represents a webhook from Discord.

    Attributes
    ----------
    id : `Snowflake`
        The ID of the webhook.
    type : `WebhookType`
        The type of the webhook.
    guild_id : `Snowflake`, optional
        The guild ID this webhook is for, if any.
    channel_id : `Snowflake`, optional
        The channel ID this webhook is for, if any.
    user : `User`, optional
        The user this webhook was created by.

        Not included when getting a webhook by its token.
    name : `str`, optional
        The default name of the webhook, if any.
    avatar : `str`, optional
        The default user avatar hash of the webhook, if any.
    token : `str`, optional
        The secure token of the webhook, if any. This is returned for `INCOMING` webhooks.
    application_id : `Snowflake`, optional
        The bot/OAuth2 application that created this webhook, if any.
    source_guild : `Guild`, optional
        The guild of the channel that this webhook is following, if any. Returned for `CHANNEL_FOLLOWER` webhooks.
    soruce_channel : `Channel`, optional
        The channel that this webhook is following, if any. Returned for `CHANNEL_FOLLOWER` webhooks.
    url : `str`, optional
        The url used for executing the webhook, if any. This is returned by the webhooks OAuth2 flow.
    """

    id: Snowflake | str = field(converter=Snowflake)
    """The ID of the webhook."""
    type: int | WebhookType = field(converter=WebhookType)
    """The type of the webhook."""
    guild_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """The guild ID this webhook is for, if any."""
    channel_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """The channel ID this webhook is for, if any."""
    user: dict | User | None = field(converter=optional_c(User), default=None)
    """The user this webhook was created by, if any."""
    name: str | None = field(default=None)
    """The default name of the webhook, if any."""
    avatar: str | None = field(default=None)
    """The default user avatar hash of the webhook, if any."""
    token: str | None = field(default=None)
    """The secure token of the webhook, if any. This is returned for `INCOMING` webhooks."""
    application_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """The bot/OAuth2 application that created this webhook, if any."""
    source_guild: dict | Guild | None = field(converter=optional_c(Guild), default=None)
    """The guild of the channel that this webhook is following, if any. Returned for `CHANNEL_FOLLOWER` webhooks."""
    source_channel: dict | Channel | None = field(converter=optional_c(Channel), default=None)
    """The channel that this webhook is following, if any. Returned for `CHANNEL_FOLLOWER` webhooks."""
    url: str | None = field(default=None)
    """The url used for executing the webhook, if any. This is returned by the webhooks OAuth2 flow."""



from ..api.http import _Route, _RouteMethod

from attrs import asdict

__all__ = ("Respondable", "Controllable", "Editable")


class Editable:
    """
    A mixin for objects with edit/delete methods.

    Methods
    -------
    `edit()` : `dict`
        Edits an object with the given path and kwargs.
    `modify()` : `dict`
        An alias of the `edit()` method.
    `delete()` : `dict | None`
        Delete an object from discord.
    """

    async def edit(self, bot: "Bot", path: str, **kwargs) -> dict:  # noqa
        """
        Edits an object with the given path and kwargs.

        Parameters
        -----------
        bot : `retux.Bot`
            The instance of the bot.
        path : `str`
            The path to the endpoint of the Discord API.
        **kwargs : `dict`
            The data to edit.

        Returns
        -------
        `dict`
            The data returned from Discord.
        """

        route = _Route(method=_RouteMethod.PATCH, path=path)
        payload = {}
        for key, value in kwargs.items():
            if hasattr(value, "__slots__"):
                payload[key] = asdict(
                    value, filter=lambda _name, _value: _name.name not in {"_bot", "bot"}
                )
            else:
                payload[key] = value

        return await bot.http.request(route, payload)

    async def modify(self, bot: "Bot", path: str, **kwargs) -> dict:  # noqa
        """An alias of the `edit()` method."""
        return await self.edit(bot, path, **kwargs)

    async def delete(self, bot: "Bot", path: str) -> None | dict:  # noqa
        """
        Delete an object from discord.

        Parameters
        ----------
        bot : `retux.Bot`
            The instance of the bot.
        path : `str`
            The path to the endpoint of the Discord API.

        Returns
        -------
        `dict | None`
            The data given from Discord, if any.
        """
        route = _Route(method=_RouteMethod.DELETE, path=path)
        return await bot.http.request(route)


class Respondable(Editable):
    """
    A mixin for objects with send/edit/delete methods.

    Methods
    -------
    respond() : `dict`
        Executes a `respond` action to Discord with the given path and keyword arguments.
    send() : `dict`
        An alias of the `respond()` method.
    """

    async def respond(self, bot: "Bot", path: str, **kwargs) -> dict:  # noqa
        """
        Executes a `respond` action to Discord with the given path and keyword arguments.

        Parameters
        ----------
        bot : `retux.Bot`
            The instance of the bot.
        path : `str`
            The path to the endpoint of the Discord API.
        **kwargs : `dict`
            The data to include in the request.

        Returns
        -------
        `dict`
            The data of the interaction response returned by Discord.
        """

        route = _Route(method=_RouteMethod.POST, path=path)

        payload = {}
        for key, value in kwargs.items():
            if hasattr(value, "__slots__"):
                payload[key] = asdict(
                    value, filter=lambda _name, _value: _name.name not in {"_bot", "bot"}
                )
            else:
                payload[key] = value

        return await bot.http.request(route, payload)

    async def send(self, bot: "Bot", path: str, **kwargs) -> dict:  # noqa
        """An alias of the `respond()` method."""
        return await self.respond(bot, path, **kwargs)


class Controllable:
    """
    A mixin for objects with create/get methods.

    Methods
    -------
    `create()` : `dict`
        Creates an object with the given path and kwargs.
    `get()` : `dict`
        Gets an object from the Discord API.
    """

    @classmethod
    async def create(cls, bot: "Bot", path: str, **kwargs) -> dict:  # noqa
        """
        Creates an object with the given path and kwargs.

        Parameters
        ----------
        bot : `retux.Bot`
            The instance of the bot.
        path : `str`
            The path to the endpoint of the Discord API.
        **kwargs : `dict`
            The data for the object creation.

        Returns
        -------
        `dict`
            The data of the object returned by Discord.
        """
        route = _Route(method=_RouteMethod.POST, path=path)

        payload = {}
        for key, value in kwargs.items():
            if hasattr(value, "__slots__"):
                payload[key] = asdict(
                    value, filter=lambda _name, _value: _name.name not in {"_bot", "bot"}
                )
            else:
                payload[key] = value

        return await bot.http.request(route, payload)

    @classmethod
    async def get(cls, bot: "Bot", path: str, **query_params: dict | None) -> dict:  # noqa
        """
        Gets an object from the Discord API.

        Parameters
        ----------
        bot : `retux.Bot`
            The instance of the bot.
        path : `str`
            The path to the endpoint of the Discord API.
        **query_params : `dict`, optional
            The data to query for when getting. For example `with_counts=True` for guilds.

        Returns
        -------
        `dict`
            The data of the object returned by Discord.
        """
        route = _Route(method=_RouteMethod.GET, path=path)
        return await bot.http.request(route, payload=query_params)
